/**************************************************************************
   STV-TOOLS: Shannon Total Variation based image processing
   Copyright (C) 2016 Remy Abergel (remy.abergel AT gmail.com),
   Lionel Moisan (Lionel.Moisan AT parisdescartes.fr).

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
***************************************************************************/

/*

  Compilation (with gcc): gcc -O3 fftspectrum.c tiffread.c tiffwrite.c -lfftw3 -lm -ltiff -o fftspectrum

  Versions :

  + v 1.0 : (09/2016) initial version (Rémy Abergel)

 */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include "fftw3.h"

#define VERSION "1.0"

#define FREE_ALL() {					\
    if(NULL != u) free(u);				\
    if(NULL != v) free(v);				\
    if(NULL != dft_u) fftw_free(dft_u);			\
    if(NULL != plan_u) fftw_destroy_plan(plan_u);	\
  }
#define ASSERT_ALLOC(cmd) if(NULL == (cmd)) { printf("Not enough memory\n"); FREE_ALL(); exit(EXIT_FAILURE); }

static void display_usage()
{
  printf("\nCompute FFT spectrum in log scale: out = log(1+|FFT(in)|).\n\n");
  printf("Usage: fftspectrum [-c] [-f] in out\n\n");
  printf("   -c  : set output min/max values equal to 0 and 255 using the appropriate affine contrast change\n");
  printf("   -f  : save output as a tiff-float instead of unsigned char tiff\n");
  printf("   in  : input image (TIFF)\n");
  printf("   out : output image (TIFF)\n\n");
}


int main(int argc, char **argv)
{
  char *filenamein = NULL,*filenameout = NULL,vflag,cflag,fflag;
  double *u = NULL,*v = NULL;
  fftw_complex *dft_u = NULL;
  fftw_plan plan_u = NULL;
  double min,max,val,c1,c2;
  int s,x,y,a,b,adr,i,err,nx,ny,half_nx,half_ny;
  vflag = cflag = fflag = 0;

  /******************************************/
  /*            arguments parser            */
  /******************************************/

  /* main parser */
  for(i=1;i<argc;i++) {
    if (strcmp(argv[i],"-c") == 0) cflag = 1;
    else if (strcmp(argv[i],"-f") == 0) fflag = 1;
    else if (NULL == filenamein) filenamein = argv[i];
    else if (NULL == filenameout) filenameout = argv[i];
    else { display_usage(); printf("Error: could not retrieve properly input arguments.\n\n"); return EXIT_FAILURE; }
  }

  /* check consistency */
  if ((NULL == filenamein) || (NULL == filenameout)) {
    display_usage();
    if (NULL == filenamein) printf("Error: input 'in' is missing\n\n");
    else if(NULL == filenameout) printf("Error: output 'out' is missing\n\n");
    return EXIT_FAILURE;
  }

  /************************************************************/
  /* Open TIFF Image & get its graylevels in double precision */
  /************************************************************/

  if(EXIT_FAILURE == read_tiff_into_double(&u,&nx,&ny,filenamein,vflag)) {
    printf("An error occured (try verbose mode using option -v to get more informations)\n");
    FREE_ALL();
    return EXIT_FAILURE;
  }

  /*********************/
  /* Memory allocation */
  /*********************/
  half_nx = nx/2; // integer division (= (int) floor((double)nx/2.))
  half_ny = ny/2; // integer division (= (int) floor((double)ny/2.))

  ASSERT_ALLOC(v = (double*) malloc(nx*ny * sizeof(double)));
  ASSERT_ALLOC(dft_u = (fftw_complex*) fftw_malloc ((1+half_nx)*ny * sizeof(fftw_complex)));
  ASSERT_ALLOC(plan_u = fftw_plan_dft_r2c_2d(ny,nx,u,dft_u,FFTW_ESTIMATE));

  /*********************************************/
  /* KERNEL: compute DFT spectrum in log scale */
  /*********************************************/
  fftw_execute(plan_u); // set dft_u = DFT(u);
  min = 0.; max = -1;
  // set v = log(1+|DFT(u)|)
  for(y=0,b=-half_ny;b<=(ny-1)/2;b++,y++)
    for(x=0,a=-half_nx;a<=(nx-1)/2;a++,x++) {
      s = (a >= 0) ? 1 : -1;
      adr = ((s*b+ny)%ny)*(1+half_nx) + (s*a+nx)%nx; // location of frequency (|a|,b) in dft_u
      val = log(1.+hypot(dft_u[adr][0],dft_u[adr][1]));
      //printf("a=%d, b=%d, val=%.10g, re=%.10g, im=%.10g, u=%.10g\n",a,b,val,dft_u[adr][0],dft_u[adr][1],u[y*nx+x]);
      v[y*nx+x] = val;
      if(min > val) min = val;
      if(max < val) max = val;
    }
  // apply contrast change (if needed)
  if(cflag && (min < max))
    for(c1=255./(max-min),c2=-c1*min,adr=0;adr<nx*ny;adr++)
      v[adr] = c1*v[adr]+c2;

  /*********************************************/
  /*               END OF KERNEL               */
  /*********************************************/




  /********************************/
  /* Write output as a TIFF Image */
  /********************************/
  if (EXIT_FAILURE == save_tiff(filenameout,v,nx,ny,8,fflag,1,vflag)) {
    printf("An error occured (try verbose mode using option -v to get more informations)\n");
    FREE_ALL();
    return EXIT_FAILURE;
  }

  /* free memory */
  FREE_ALL();
}
