/**************************************************************************
   STV-TOOLS: Shannon Total Variation based image processing
   Copyright (C) 2016 Remy Abergel (remy.abergel AT gmail.com),
   Lionel Moisan (Lionel.Moisan AT parisdescartes.fr).

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
***************************************************************************/

/*

  Compilation (with gcc): gcc -O3 add_images.c tiffread.c tiffwrite.c -lm -ltiff -o add_images

  References :

  [1] R. Abergel and L. Moisan, ``The Shannon Total Variation'', preprint MAP5 n° 2016-19, July 2016.
  [2] L. Moisan. How to discretize the total variation of an image? Proceedings in Applied Mathematics and Mechanics, 2007.

  Versions :

  + v 1.0 : (09/2016) initial version (Rémy Abergel)

 */

#include <string.h>
#include <math.h>
#include "fftw3.h"

/* compute_sgrad : Shannon gradient (gx,gy) = Grad_n(u) */
void compute_sgrad(dft_u,dft_gx,dft_gy,plan_u,iplan_gx,iplan_gy,nx,ny,Nx,Ny)
     fftw_complex *dft_u,*dft_gx,*dft_gy;
     fftw_plan plan_u,iplan_gx,iplan_gy;
     int nx,ny,Nx,Ny;
{
  int a,b,adr1,adr2,half_nx,half_Nx,half_ny;
  double cof,re,im,dft_nrm1,dft_nrm2;
  char nx_is_even,ny_is_even;

  /* initialize local variables */
  nx_is_even = !(nx%2);
  ny_is_even = !(ny%2);
  half_nx = nx/2; // integer division = floor((double)nx/2.)
  half_ny = ny/2; // integer division = floor((double)ny/2.)
  half_Nx = Nx/2; // integer division = floor((double)Nx/2.)
  dft_nrm1 = dft_nrm2 = 2.*M_PI/((double)nx*(double)ny);
  dft_nrm1 /= (double)nx;
  dft_nrm2 /= (double)ny;

  /* compute the DFT of u into dft_u */
  fftw_execute(plan_u);

  /* set dft_gx and dft_gy equal to zero (see remark below) */
  memset(dft_gx,0,(1+half_Nx)*Ny*sizeof(fftw_complex));
  memset(dft_gy,0,(1+half_Nx)*Ny*sizeof(fftw_complex));

  /* fill dft_gx and dft_gy with 1./(Nx*Ny) * DFT of the Shannon gradient of u */
  for(b=-half_ny;b<=half_ny;b++) {
    for(a=0;a<=half_nx;a++) {
      adr1 = ((b+ny)%ny)*(1+half_nx)+a; // = location of freq. (a,b) in dft_u
      adr2 = ((b+Ny)%Ny)*(1+half_Nx)+a; // = location of freq. (a,b) in dft_gx & dft_gy
      re   = dft_u[adr1][0];
      im   = dft_u[adr1][1];
      cof = ((nx_is_even && (a==half_nx)) ? .5 : 1.) * ((ny_is_even && (fabs(b)==half_ny)) ? .5 : 1.);
      dft_gx[adr2][0] = - dft_nrm1 * a * im * cof;
      dft_gx[adr2][1] =   dft_nrm1 * a * re * cof;
      dft_gy[adr2][0] = - dft_nrm2 * b * im * cof;
      dft_gy[adr2][1] =   dft_nrm2 * b * re * cof;
    }
  }

  /* compute gx and gy as the inverse DFTs of dft_gx and dft_gy */
  // BEWARE: the following commands will change the values of dft_gx and dft_gy (see FFTW3 documentation)
  fftw_execute(iplan_gx);
  fftw_execute(iplan_gy);
}


/* compute_sdiv : Shannon divergence, div = div_n(px,py) */
void compute_sdiv(dft_div,dft_px,dft_py,plan_px,plan_py,iplan_div,nx,ny,Nx,Ny)
     fftw_complex *dft_div,*dft_px,*dft_py;
     fftw_plan plan_px,plan_py,iplan_div;
     int nx,ny,Nx,Ny;
{
  int a,b,adr1,adr2,adr3,a0,b0,half_nx,half_ny,half_Nx;
  double dft_nrm1,dft_nrm2,hpx_re,hpx_im,hpy_re,hpy_im;
  char nx_is_even,ny_is_even;

  /* initialize local variables */
  nx_is_even = !(nx%2);
  ny_is_even = !(ny%2);
  dft_nrm1 = dft_nrm2 = 2.*M_PI/((double)nx*(double)ny);
  dft_nrm1 /= (double)nx;
  dft_nrm2 /= (double)ny;
  a0 = (nx-1)/2; // integer division = floor(((double)nx-1.)/2.)
  b0 = (ny-1)/2; // integer division = floor(((double)ny-1.)/2.)
  half_nx = nx/2; // integer division = floor((double)nx/2.)
  half_ny = ny/2; // integer division = floor((double)ny/2.)
  half_Nx = Nx/2; // integer division = floor((double)Nx/2.)

  /* initialize local variables */
  a0 = (nx-1)/2; // integer division = floor(((double)nx-1.)/2.)
  b0 = (ny-1)/2; // integer division = floor(((double)ny-1.)/2.)

  /* set dft_px = DFT(px) and dft_py = DFT(py) */
  fftw_execute(plan_px);
  fftw_execute(plan_py);

  /* first pass: deal with Shannon frequencies {(a,b), |a|<0.5*nx and |b|<0.5*ny} */
  for(b=-b0;b<=b0;b++)
    for(a=0;a<=a0;a++) {
      adr1 = ((b+ny)%ny)*(1+half_nx)+a; // = location of freq. (a,b) in dft_div
      adr2 = ((b+Ny)%Ny)*(1+half_Nx)+a;  // = location of freq. (a,b) in dft_px & dft_py
      dft_div[adr1][0] = - dft_nrm1*a*dft_px[adr2][1] - dft_nrm2*b*dft_py[adr2][1];
      dft_div[adr1][1] = dft_nrm1*a*dft_px[adr2][0] + dft_nrm2*b*dft_py[adr2][0];
    }

  /* second pass: if nx is even, deal with Non-Shannon frequencies {(nx/2,b), |b| < 0.5*ny} */
  if(nx_is_even) {
    // compute DFT(div)(-half_nx,b) for |b|<0.5*ny
    // using (for a=-half_nx)
    //
    // DFT(div)(a,b) =   2*i*pi*a/nx*(DFT(px)(a,b)-DFT(px)(-a,b))/2
    //                 + 2*i*pi*b/ny*(DFT(py)(a,b)+DFT(py)(-a,b))/2
    //
    for(a=-half_nx,b=-b0;b<=b0;b++) {
      adr1 = ((b+ny)%ny)*(1+half_nx)+half_nx; // = location of freq. (-half_nx,b) in dft_div (same location as freq. (half_nx,b))
      adr2 = ((b+Ny)%Ny)*(1+half_Nx)+half_nx; // = location of freq. (half_nx,b) in dft_px & dft_py
      adr3 = ((-b+Ny)%Ny)*(1+half_Nx)+half_nx; // = location of freq. (half_nx,-b) in dft_px & dft_py
      hpx_re = 0.5*(dft_px[adr3][0]-dft_px[adr2][0]);
      hpx_im = 0.5*(-dft_px[adr3][1]-dft_px[adr2][1]); // recall that DFT(px)(-half_nx,b) = Conj(DFT(px)(half_nx,-b))), same remark for dft_py.
      hpy_re = 0.5*(dft_py[adr3][0]+dft_py[adr2][0]);
      hpy_im = 0.5*(-dft_py[adr3][1]+dft_py[adr2][1]); // recall that DFT(px)(-half_nx,b) = Conj(DFT(px)(half_nx,-b))), same remark for dft_py.
      dft_div[adr1][0] = -(dft_nrm1*a*hpx_im + dft_nrm2*b*hpy_im);
      dft_div[adr1][1] =  (dft_nrm1*a*hpx_re + dft_nrm2*b*hpy_re);
    }
  }

  /* third pass: if ny is even, deal with Non-Shannon frequencies {(a,ny/2), |a| < 0.5*nx} */
  if(ny_is_even) {
    // compute DFT(div)(a,-half_ny) for |a|<0.5*nx
    // using (for b=-half_ny)
    //
    // DFT(div)(a,b) =   2*i*pi*a/nx*(DFT(px)(a,b)+DFT(px)(a,-b))/2
    //                 + 2*i*pi*b/ny*(DFT(py)(a,b)-DFT(py)(a,-b))/2
    //
    for(b=-half_ny,a=0;a<=a0;a++) {
      adr1 = half_ny*(1+half_nx)+a; // = location of freq. (a,-half_ny) in dft_div
      adr2 = ((b+Ny)%Ny)*(1+half_Nx)+a;  // = location of freq. (a,b) in dft_px & dft_py
      adr3 = ((-b+Ny)%Ny)*(1+half_Nx)+a;  // = location of freq. (a,-b) in dft_px & dft_py
      hpx_re = 0.5*(dft_px[adr3][0]+dft_px[adr2][0]);
      hpx_im = 0.5*(dft_px[adr3][1]+dft_px[adr2][1]);
      hpy_re = 0.5*(dft_py[adr2][0]-dft_py[adr3][0]);
      hpy_im = 0.5*(dft_py[adr2][1]-dft_py[adr3][1]);
      dft_div[adr1][0] = -(dft_nrm1*a*hpx_im + dft_nrm2*b*hpy_im);
      dft_div[adr1][1] =  (dft_nrm1*a*hpx_re + dft_nrm2*b*hpy_re);
    }
  }

  /* last pass: if nx and ny are both even, deal with the Non-Shannon frequencies (a,b) = (+/- nx/2, +/- ny/2) */
  if(nx_is_even && ny_is_even) {
    // compute DFT(div)(-half_nx,-half_ny)
    // using (for a=-half_nx, b=-half_ny)
    //
    // DFT(div)(a,b) =   2*i*pi*a/nx*(DFT(px)(a,b)+DFT(px)(a,-b)-DFT(px)(-a,b)-DFT(px)(-a,-b))/4
    //                 + 2*i*pi*b/ny*(DFT(py)(a,b)-DFT(py)(a,-b)+DFT(py)(-a,b)-DFT(py)(-a,-b))/4
    //
    a = -half_nx;
    b = -half_ny;
    adr1 = half_ny*(1+half_nx)+half_nx; // = location of freq. (-half_nx,-half_ny) in dft_div (same location as freq. (half_nx,half_ny))
    adr2 = ((b+Ny)%Ny)*(1+half_Nx) + half_nx; // = location of freq. (half_nx,-half_ny) in dft_px & dft_py
    adr3 = ((-b+Ny)%Ny)*(1+half_Nx) + half_nx; // = location of freq. (half_nx,half_ny) in dft_px & dft_py
    /*
    hpx_re = 0.; // = 0.25*(dft_px[adr3][0]+dft_px[adr2][0]-dft_px[adr2][0]-dft_px[adr3][0])
    hpy_re = 0.; // = 0.25*(dft_py[adr3][0]-dft_py[adr2][0]+dft_py[adr2][0]-dft_py[adr3][0]) = 0
    */
    hpx_im = -0.5*(dft_px[adr3][1]+dft_px[adr2][1]);
    hpy_im = 0.5*(-dft_py[adr3][1]+dft_py[adr2][1]);
    dft_div[adr1][0] = -(dft_nrm1*a*hpx_im + dft_nrm2*b*hpy_im);
    dft_div[adr1][1] = 0.; // = dft_nrm1*a*hpx_re + dft_nrm2*b*hpy_re = 0
  }

  /* set div = inverse DFT of dft_div */
  fftw_execute(iplan_div);
}
