/************************************************************************************

   STV-TOOLS: Shannon Total Variation based image processing Copyright (C) 2016
   Remy Abergel (remy.abergel AT gmail.com), Lionel Moisan (Lionel.Moisan AT
   parisdescartes.fr).

   This program is free software: you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free Software
   Foundation, either version 3 of the License, or (at your option) any later
   version.

   This program is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
   FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
   details.

   You should have received a copy of the GNU General Public License along with
   this program.  If not, see <http://www.gnu.org/licenses/>.

                           IMPORTANT NOTE

   Part of this code is inspired from the source file CImg.h of the CImg library
   (David Tschumperlé - GREYC UMR CNRS 6072, Image team).

   The CImg software is governed either by the CeCILL or CeCILL-C license under
   French law and abiding by the rules of distribution of free software. You can
   use, modify and or redistribute the software under the terms of the CeCILL or
   CeCILL-C licenses as circulated by CEA, CNRS and INRIA at the following URL:
   "http://www.cecill.info".

   As a counterpart to the access to the source code and rights to copy, modify
   and redistribute granted by the license, users are provided only with a
   limited warranty and the software's author, the holder of the economic
   rights, and the successive licensors have only limited liability.

   In this respect, the user's attention is drawn to the risks associated with
   loading, using, modifying and/or developing or reproducing the software by
   the user in light of its specific status of free software, that may mean that
   it is complicated to manipulate, and that also therefore means that it is
   reserved for developers and experienced professionals having in-depth
   computer knowledge. Users are therefore encouraged to load and test the
   software's suitability as regards their requirements in conditions enabling
   the security of their systems and/or data to be ensured and, more generally,
   to use and operate it in the same conditions as regards security.

   The fact that you are presently reading this means that you have had
   knowledge of the CeCILL and CeCILL-C licenses and that you accept its terms.

*************************************************************************************/

/*

  Versions :

  + v 1.0 : (09/2016) initial version (Rémy Abergel)

  Note about Supported TIFF files:
  ================================

  Are supported: Tiled or Striped, Contiguous or Separate Planar Config,
  monopage & monochannel images, with (u)int(8,16,32) or float precision.

  Compression: only non-compressed or LZW-compressed TIFF are supported.

 */

#include <stdlib.h>
#include <stdio.h>
#include <tiffio.h>

/* preprocessor functions */
#define MIN(a,b)        (((a)>(b)) ? (b) : (a))
#define ASSERT_ALLOC(cmd)	if(NULL == (cmd)) \
				{ \
					printf("Not enough memory\n"); \
					return EXIT_FAILURE; \
				}
#define RAISE_ERROR(message) 	TIFFClose(tiffp); printf("%s",message);

/* prototypes */
int read_tiff_into_double(double**,int*,int*,char*,char);

static int _load_tiff_tiled_contig_uint8(double***,TIFF*,const uint32,const uint32);
static int _load_tiff_tiled_contig_int8(double***,TIFF*,const uint32,const uint32);
static int _load_tiff_tiled_contig_uint16(double***,TIFF*,const uint32,const uint32);
static int _load_tiff_tiled_contig_int16(double***,TIFF*,const uint32,const uint32);
static int _load_tiff_tiled_contig_uint32(double***,TIFF*,const uint32,const uint32);
static int _load_tiff_tiled_contig_int32(double***,TIFF*,const uint32,const uint32);
static int _load_tiff_tiled_contig_float(double***,TIFF*,const uint32,const uint32);

static int _load_tiff_tiled_separate_uint8(double***,TIFF*,const uint32,const uint32);
static int _load_tiff_tiled_separate_int8(double***,TIFF*,const uint32,const uint32);
static int _load_tiff_tiled_separate_uint16(double***,TIFF*,const uint32,const uint32);
static int _load_tiff_tiled_separate_int16(double***,TIFF*,const uint32,const uint32);
static int _load_tiff_tiled_separate_uint32(double***,TIFF*,const uint32,const uint32);
static int _load_tiff_tiled_separate_int32(double***,TIFF*,const uint32,const uint32);
static int _load_tiff_tiled_separate_float(double***,TIFF*,const uint32,const uint32);

static int _load_tiff_striped_contig_uint8(double***,TIFF*);
static int _load_tiff_striped_contig_int8(double***,TIFF*);
static int _load_tiff_striped_contig_uint16(double***,TIFF*);
static int _load_tiff_striped_contig_int16(double***,TIFF*);
static int _load_tiff_striped_contig_uint32(double***,TIFF*);
static int _load_tiff_striped_contig_int32(double***,TIFF*);
static int _load_tiff_striped_contig_float(double***,TIFF*);

static int _load_tiff_striped_separate_uint8(double***,TIFF*);
static int _load_tiff_striped_separate_int8(double***,TIFF*);
static int _load_tiff_striped_separate_uint16(double***,TIFF*);
static int _load_tiff_striped_separate_int16(double***,TIFF*);
static int _load_tiff_striped_separate_uint32(double***,TIFF*);
static int _load_tiff_striped_separate_int32(double***,TIFF*);
static int _load_tiff_striped_separate_float(double***,TIFF*);

/* globale variables */
static char *filename;
static uint16 bitspersample,sampleformat = SAMPLEFORMAT_UINT;
static uint32 nx,ny;

/* Main loading module:
 *
 *  + open a TIFF File,
 *
 *  + retrieve dimensions, number of channels and datatype for each directory
 *  + check consistency (dimensions, number of channel and datatype must be the
 *    sames for all directories)
 *  + allocate memory for pixmap
 *  + load the first directory (= directory 0), using the appropriate
 *    _load_tiff_* module.
 *
 * This function returns EXIT_SUCCESS or EXIT_FAILURE integer.
 *
 * */
int read_tiff_into_double(double **graylevels, int *width, int *height, char *filename, char verbose)
{

  TIFF *tiffp;
  unsigned int npages = 0;
  uint16 compression_type,samplesperpixel,photo,config;
  uint32 x,y,adr,tw,th;

  if(verbose) {
    printf("load input TIFF image '%s': ",filename);
    fflush(stdout);
  }

  /* open tiff file & read directories */
  tiffp = TIFFOpen(filename,"r");
  if (!tiffp) {
    printf("Failed to open TIFF File '%s'\n",filename);
    return EXIT_FAILURE;
  }
  else {
    if (verbose) { printf("\33[2K\rLoad input TIFF image '%s': in progress",filename); fflush(stdout); }
    /* retrieve basic image properties */
    do ++npages; while (TIFFReadDirectory(tiffp));
    if (npages == 0) {
      if (verbose) { printf("\33[2K\rLoad input TIFF image '%s': failure.\n",filename); fflush(stdout); }
      RAISE_ERROR("No image detected (number of TIFF page = 0)\n");
      return EXIT_FAILURE;
    }
    if (npages > 1) {
      if (verbose) { printf("\33[2K\rLoad input TIFF image '%s': failure.\n",filename); fflush(stdout); }
      RAISE_ERROR("Multipage TIFF files are not unsupported\n");
      return EXIT_FAILURE;
    }
    TIFFSetDirectory(tiffp,0);
    TIFFGetField(tiffp,TIFFTAG_IMAGEWIDTH,&nx);
    TIFFGetField(tiffp,TIFFTAG_IMAGELENGTH,&ny);
    TIFFGetFieldDefaulted(tiffp,TIFFTAG_SAMPLESPERPIXEL,&samplesperpixel);
    if (samplesperpixel != 1) {
      if (verbose) { printf("\33[2K\rLoad input TIFF image '%s': failure.\n",filename); fflush(stdout); }
      RAISE_ERROR("Multichannel TIFF files are not unsupported. Please, convert your TIFF image into a single-channel grayscaled image.\n");
      return EXIT_FAILURE;
    }
    TIFFGetFieldDefaulted(tiffp,TIFFTAG_SAMPLEFORMAT,&sampleformat);
    TIFFGetFieldDefaulted(tiffp,TIFFTAG_BITSPERSAMPLE,&bitspersample);
    TIFFGetFieldDefaulted(tiffp,TIFFTAG_COMPRESSION,&compression_type);
    if((compression_type != COMPRESSION_NONE)&&(compression_type != COMPRESSION_LZW)) {
      if (verbose) { printf("\33[2K\rLoad input TIFF image '%s': failure.\n",filename); fflush(stdout); }
      printf("Unsupported TIFF compression type (%d). This reader only supports Uncompressed (compression type = %d) and LZW-compressed (compression_type =%d) images.",compression_type,COMPRESSION_NONE,COMPRESSION_LZW);
      RAISE_ERROR("Unsupported TIFF compression (only non-compressed or LZW compressed TIFF images are supported).");
      return EXIT_FAILURE;
    }

    /* check sample format */
    if (!((sampleformat==SAMPLEFORMAT_UINT)||(sampleformat==SAMPLEFORMAT_INT)||(sampleformat==SAMPLEFORMAT_IEEEFP)||(sampleformat==SAMPLEFORMAT_VOID))) {
      if (verbose) { printf("\33[2K\rLoad input TIFF image '%s': failure.\n",filename); fflush(stdout); }
      printf("Unsupported TIFF sample format (%d), this reader only supports (u)int and float images (sample format = %d, %d or %d).\nNote also that undefined data format are accepted (sample format=%d) but they are loaded as unsigned integer.",sampleformat,SAMPLEFORMAT_UINT,SAMPLEFORMAT_INT,SAMPLEFORMAT_IEEEFP,SAMPLEFORMAT_VOID);
      TIFFClose(tiffp);
      return EXIT_FAILURE;
    }
    if(sampleformat==SAMPLEFORMAT_VOID) {
      //printf("Undefined sample format, let us try to load the image with unsigned integer precision.\n");
      sampleformat = SAMPLEFORMAT_UINT;
    }

    /* memory allocation */
    ASSERT_ALLOC((*graylevels) = (double*) malloc (nx*ny*sizeof(double)));

    /* retrieve graylevels (store them in double precision) */
    if (!TIFFSetDirectory(tiffp,0)) {
      if (verbose) { printf("\33[2K\rLoad input TIFF image '%s': failure.\n",filename); fflush(stdout); }
      RAISE_ERROR("Could not reach TIFF directory, command TIFFSetDirectory(tiffp,0) failed");
    }
    TIFFGetField(tiffp,TIFFTAG_PLANARCONFIG,&config);
    TIFFGetField(tiffp,TIFFTAG_PHOTOMETRIC,&photo);
    if (TIFFIsTiled(tiffp)) { // tiled TIFF case
      TIFFGetField(tiffp,TIFFTAG_TILEWIDTH,&tw);
      TIFFGetField(tiffp,TIFFTAG_TILELENGTH,&th);
      if (config==PLANARCONFIG_CONTIG) { // tiled contiguous planar config.
	switch (bitspersample) {
	case 8 :
	  if (sampleformat==SAMPLEFORMAT_UINT) _load_tiff_tiled_contig_uint8(&graylevels,tiffp,tw,th);
	  else _load_tiff_tiled_contig_int8(&graylevels,tiffp,tw,th);
	  break;
	case 16 :
	  if (sampleformat==SAMPLEFORMAT_UINT) _load_tiff_tiled_contig_uint16(&graylevels,tiffp,tw,th);
	  else _load_tiff_tiled_contig_int16(&graylevels,tiffp,tw,th);
	  break;
	case 32 :
	  if (sampleformat==SAMPLEFORMAT_UINT) _load_tiff_tiled_contig_uint32(&graylevels,tiffp,tw,th);
	  else if (sampleformat==SAMPLEFORMAT_INT) _load_tiff_tiled_contig_int32(&graylevels,tiffp,tw,th);
	  else _load_tiff_tiled_contig_float(&graylevels,tiffp,tw,th);
	  break;
	default :
	  if (verbose) { printf("\33[2K\rLoad input TIFF image '%s': failure.\n",filename); fflush(stdout); }
	  return EXIT_FAILURE;
	}
      }
      else { // tiled separate (=non-contiguous) planar config.
	switch (bitspersample) {
	case 8 :
	  if (sampleformat==SAMPLEFORMAT_UINT) _load_tiff_tiled_separate_uint8(&graylevels,tiffp,tw,th);
	  else _load_tiff_tiled_separate_int8(&graylevels,tiffp,tw,th);
	  break;
	case 16 :
	  if (sampleformat==SAMPLEFORMAT_UINT) _load_tiff_tiled_separate_uint16(&graylevels,tiffp,tw,th);
	  else _load_tiff_tiled_separate_int16(&graylevels,tiffp,tw,th);
	  break;
	case 32 :
	  if (sampleformat==SAMPLEFORMAT_UINT) _load_tiff_tiled_separate_uint32(&graylevels,tiffp,tw,th);
	  else if (sampleformat==SAMPLEFORMAT_INT) _load_tiff_tiled_separate_int32(&graylevels,tiffp,tw,th);
	  else _load_tiff_tiled_separate_float(&graylevels,tiffp,tw,th);
	  break;
	default :
	  if (verbose) { printf("\33[2K\rLoad input TIFF image '%s': failure.\n",filename); fflush(stdout); }
	  return EXIT_FAILURE;
	}
      }
    }
    else { // striped (=non-tiled) TIFF case
      if (config==PLANARCONFIG_CONTIG) { // striped (=non-tiled) contiguous planar config
	switch (bitspersample) {
	case 8 :
	  if (sampleformat==SAMPLEFORMAT_UINT) _load_tiff_striped_contig_uint8(&graylevels,tiffp);
	  else _load_tiff_striped_contig_int8(&graylevels,tiffp);
	  break;
	case 16 :
	  if (sampleformat==SAMPLEFORMAT_UINT) _load_tiff_striped_contig_uint16(&graylevels,tiffp);
	  else _load_tiff_striped_contig_int16(&graylevels,tiffp);
	  break;
	case 32 :
	  if (sampleformat==SAMPLEFORMAT_UINT) _load_tiff_striped_contig_uint32(&graylevels,tiffp);
	  else if (sampleformat==SAMPLEFORMAT_INT) _load_tiff_striped_contig_int32(&graylevels,tiffp);
	  else _load_tiff_striped_contig_float(&graylevels,tiffp);
	  break;
	default :
	  if (verbose) { printf("\33[2K\rLoad input TIFF image '%s': failure.\n",filename); fflush(stdout); }
	  return EXIT_FAILURE;
	}
      }
      else { // striped (=non-tiled) separate (=non-contiguous) planar config
	switch (bitspersample) {
	case 8 :
	  if (sampleformat==SAMPLEFORMAT_UINT) _load_tiff_striped_separate_uint8(&graylevels,tiffp);
	  else _load_tiff_striped_separate_int8(&graylevels,tiffp);
	  break;
	case 16 :
	  if (sampleformat==SAMPLEFORMAT_UINT) _load_tiff_striped_separate_uint16(&graylevels,tiffp);
	  else _load_tiff_striped_separate_int16(&graylevels,tiffp);
	  break;
	case 32 :
	  if (sampleformat==SAMPLEFORMAT_UINT) _load_tiff_striped_separate_uint32(&graylevels,tiffp);
	  else if (sampleformat==SAMPLEFORMAT_INT) _load_tiff_striped_separate_int32(&graylevels,tiffp);
	  else _load_tiff_striped_separate_float(&graylevels,tiffp);
	  break;
	default :
	  if (verbose) { printf("\33[2K\rLoad input TIFF image '%s': failure.\n",filename); fflush(stdout); }
	  return EXIT_FAILURE;
	}
      }
    }
  }

  *width = nx; *height = ny;
  TIFFClose(tiffp);

  /* deal with verbose mode */
  if(verbose) {
    printf("\33[2K\rload input TIFF image '%s': done\n",filename);
    fflush(stdout);
    printf(" -> width x height = %d x %d\n",nx,ny);
    printf(" -> precision = %d bits per pixel %s\n",bitspersample,sampleformat==SAMPLEFORMAT_UINT ? "(unsigned)" : "(signed)");
    if (sampleformat==SAMPLEFORMAT_IEEEFP) printf(" -> this image is a tiff-float image\n");
  }

  return EXIT_SUCCESS;
}





/***************************************************************************************
 * Tiled Contiguous Planar Config = TIFF loaders for:
 *
 * 	- tiled (=non-striped) images, having contiguous planar config.
 * 	- supported types: uint8, int8, uint16, int16, uint32, int32 and float.
 * 	- multi-channel images supported
 *
 ***************************************************************************************/

static int _load_tiff_tiled_contig_uint8(double ***graylevels, TIFF *tiffp, const uint32 tw, const uint32 th)
{
  unsigned int row,col,rr,cc;
  uint8 *buf = (uint8*)_TIFFmalloc(TIFFTileSize(tiffp));
  if (buf) {
    for (row = 0; row<ny; row+=th)
      for (col = 0; col<nx; col+=tw) {
	if (TIFFReadTile(tiffp,buf,col,row,0,0)<0) { _TIFFfree(buf); RAISE_ERROR("Invalid tile found into the TIFF image\n"); }
	uint8 *ptr = buf;
	for (rr = row; rr<MIN(row+th,ny); ++rr)
	  for (cc = col; cc<MIN(col+tw,nx); ++cc)
	    (**graylevels)[rr*nx+cc] = (double) ptr[(rr-row)*th + (cc-col)];
      }
    _TIFFfree(buf);
  }
  return EXIT_SUCCESS;
}

static int _load_tiff_tiled_contig_int8(double ***graylevels, TIFF *tiffp, const uint32 tw, const uint32 th)
{
  unsigned int row,col,rr,cc;
  int8 *buf = (int8*)_TIFFmalloc(TIFFTileSize(tiffp));
  if (buf) {
    for (row = 0; row<ny; row+=th)
      for (col = 0; col<nx; col+=tw) {
	if (TIFFReadTile(tiffp,buf,col,row,0,0)<0) { _TIFFfree(buf); RAISE_ERROR("Invalid tile found into the TIFF image\n"); }
	int8 *ptr = buf;
	for (rr = row; rr<MIN(row+th,ny); ++rr)
	  for (cc = col; cc<MIN(col+tw,nx); ++cc)
	    (**graylevels)[rr*nx+cc] = (double) ptr[(rr-row)*th + (cc-col)];
      }
    _TIFFfree(buf);
  }
  return EXIT_SUCCESS;
}


static int _load_tiff_tiled_contig_uint16(double ***graylevels, TIFF *tiffp, const uint32 tw, const uint32 th)
{
  unsigned int row,col,rr,cc;
  uint16 *buf = (uint16*)_TIFFmalloc(TIFFTileSize(tiffp));
  if (buf) {
    for (row = 0; row<ny; row+=th)
      for (col = 0; col<nx; col+=tw) {
	if (TIFFReadTile(tiffp,buf,col,row,0,0)<0) { _TIFFfree(buf); RAISE_ERROR("Invalid tile found into the TIFF image\n"); }
	uint16 *ptr = buf;
	for (rr = row; rr<MIN(row+th,ny); ++rr)
	  for (cc = col; cc<MIN(col+tw,nx); ++cc)
	    (**graylevels)[rr*nx+cc] = (double) ptr[(rr-row)*th + (cc-col)];
      }
    _TIFFfree(buf);
  }
  return EXIT_SUCCESS;
}

static int _load_tiff_tiled_contig_int16(double ***graylevels, TIFF *tiffp, const uint32 tw, const uint32 th)
{
  unsigned int row,col,rr,cc;
  int16 *buf = (int16*)_TIFFmalloc(TIFFTileSize(tiffp));
  if (buf) {
    for (row = 0; row<ny; row+=th)
      for (col = 0; col<nx; col+=tw) {
	if (TIFFReadTile(tiffp,buf,col,row,0,0)<0) { _TIFFfree(buf); RAISE_ERROR("Invalid tile found into the TIFF image\n"); }
	int16 *ptr = buf;
	for (rr = row; rr<MIN(row+th,ny); ++rr)
	  for (cc = col; cc<MIN(col+tw,nx); ++cc)
	    (**graylevels)[rr*nx+cc] = (double) ptr[(rr-row)*th + (cc-col)];
      }
    _TIFFfree(buf);
  }
  return EXIT_SUCCESS;
}

static int _load_tiff_tiled_contig_uint32(double ***graylevels, TIFF *tiffp, const uint32 tw, const uint32 th)
{
  unsigned int row,col,rr,cc;
  uint32 *buf = (uint32*)_TIFFmalloc(TIFFTileSize(tiffp));
  if (buf) {
    for (row = 0; row<ny; row+=th)
      for (col = 0; col<nx; col+=tw) {
	if (TIFFReadTile(tiffp,buf,col,row,0,0)<0) { _TIFFfree(buf); RAISE_ERROR("Invalid tile found into the TIFF image\n"); }
	uint32 *ptr = buf;
	for (rr = row; rr<MIN(row+th,ny); ++rr)
	  for (cc = col; cc<MIN(col+tw,nx); ++cc)
	    (**graylevels)[rr*nx+cc] = (double) ptr[(rr-row)*th + (cc-col)];
      }
    _TIFFfree(buf);
  }
  return EXIT_SUCCESS;
}

static int _load_tiff_tiled_contig_int32(double ***graylevels, TIFF *tiffp, const uint32 tw, const uint32 th)
{
  unsigned int row,col,rr,cc;
  int32 *buf = (int32*)_TIFFmalloc(TIFFTileSize(tiffp));
  if (buf) {
    for (row = 0; row<ny; row+=th)
      for (col = 0; col<nx; col+=tw) {
	if (TIFFReadTile(tiffp,buf,col,row,0,0)<0) { _TIFFfree(buf); RAISE_ERROR("Invalid tile found into the TIFF image\n"); }
	int32 *ptr = buf;
	for (rr = row; rr<MIN(row+th,ny); ++rr)
	  for (cc = col; cc<MIN(col+tw,nx); ++cc)
	    (**graylevels)[rr*nx+cc] = (double) ptr[(rr-row)*th + (cc-col)];
      }
    _TIFFfree(buf);
  }
  return EXIT_SUCCESS;
}

static int _load_tiff_tiled_contig_float(double ***graylevels, TIFF *tiffp, const uint32 tw, const uint32 th)
{
  unsigned int row,col,rr,cc;
  float *buf = (float*)_TIFFmalloc(TIFFTileSize(tiffp));
  if (buf) {
    for (row = 0; row<ny; row+=th)
      for (col = 0; col<nx; col+=tw) {
	if (TIFFReadTile(tiffp,buf,col,row,0,0)<0) { _TIFFfree(buf); RAISE_ERROR("Invalid tile found into the TIFF image\n"); }
	float *ptr = buf;
	for (rr = row; rr<MIN(row+th,ny); ++rr)
	  for (cc = col; cc<MIN(col+tw,nx); ++cc)
	    (**graylevels)[rr*nx+cc] = (double) ptr[(rr-row)*th + (cc-col)];
      }
    _TIFFfree(buf);
  }
  return EXIT_SUCCESS;
}


/***************************************************************************************
 * Tiled Separate Planar Config = TIFF loaders for:
 *
 * 	- tiled (=non-striped) images, having separate (=non-contiguous) planar config.
 * 	- supported types: uint8, int8, uint16, int16, uint32, int32 and float.
 * 	- multi-channel images supported
 *
 ***************************************************************************************/

static int _load_tiff_tiled_separate_uint8(double ***graylevels, TIFF *tiffp, const uint32 tw, const uint32 th)
{
  unsigned int row,col,rr,cc;
  uint8 *buf = (uint8*)_TIFFmalloc(TIFFTileSize(tiffp));
  if (buf) {
    for (row = 0; row<ny; row+=th)
      for (col = 0; col<nx; col+=tw) {
	if (TIFFReadTile(tiffp,buf,col,row,0,0)<0) {_TIFFfree(buf); RAISE_ERROR("Invalid tile found into the TIFF image\n"); }
	uint8 *ptr = buf;
	for (rr = row; rr<MIN(row+th,ny); ++rr)
	  for (cc = col; cc<MIN(col+tw,nx); ++cc)
	    (**graylevels)[rr*nx+cc] = (double)(*(ptr++));
      }
    _TIFFfree(buf);
  }
  return EXIT_SUCCESS;
}

static int _load_tiff_tiled_separate_int8(double ***graylevels, TIFF *tiffp, const uint32 tw, const uint32 th)
{
  unsigned int row,col,rr,cc;
  int8 *buf = (int8*)_TIFFmalloc(TIFFTileSize(tiffp));
  if (buf) {
    for (row = 0; row<ny; row+=th)
      for (col = 0; col<nx; col+=tw) {
	if (TIFFReadTile(tiffp,buf,col,row,0,0)<0) {_TIFFfree(buf); RAISE_ERROR("Invalid tile found into the TIFF image\n"); }
	int8 *ptr = buf;
	for (rr = row; rr<MIN(row+th,ny); ++rr)
	  for (cc = col; cc<MIN(col+tw,nx); ++cc)
	    (**graylevels)[rr*nx+cc] = (double)(*(ptr++));
      }
    _TIFFfree(buf);
  }
  return EXIT_SUCCESS;
}

static int _load_tiff_tiled_separate_uint16(double ***graylevels, TIFF *tiffp, const uint32 tw, const uint32 th)
{
  unsigned int row,col,rr,cc;
  uint16 *buf = (uint16*)_TIFFmalloc(TIFFTileSize(tiffp));
  if (buf) {
    for (row = 0; row<ny; row+=th)
      for (col = 0; col<nx; col+=tw) {
	if (TIFFReadTile(tiffp,buf,col,row,0,0)<0) {_TIFFfree(buf); RAISE_ERROR("Invalid tile found into the TIFF image\n"); }
	uint16 *ptr = buf;
	for (rr = row; rr<MIN(row+th,ny); ++rr)
	  for (cc = col; cc<MIN(col+tw,nx); ++cc)
	    (**graylevels)[rr*nx+cc] = (double)(*(ptr++));
      }
    _TIFFfree(buf);
  }
  return EXIT_SUCCESS;
}

static int _load_tiff_tiled_separate_int16(double ***graylevels, TIFF *tiffp, const uint32 tw, const uint32 th)
{
  unsigned int row,col,rr,cc;
  int16 *buf = (int16*)_TIFFmalloc(TIFFTileSize(tiffp));
  if (buf) {
    for (row = 0; row<ny; row+=th)
      for (col = 0; col<nx; col+=tw) {
	if (TIFFReadTile(tiffp,buf,col,row,0,0)<0) {_TIFFfree(buf); RAISE_ERROR("Invalid tile found into the TIFF image\n"); }
	int16 *ptr = buf;
	for (rr = row; rr<MIN(row+th,ny); ++rr)
	  for (cc = col; cc<MIN(col+tw,nx); ++cc)
	    (**graylevels)[rr*nx+cc] = (double)(*(ptr++));
      }
    _TIFFfree(buf);
  }
  return EXIT_SUCCESS;
}

static int _load_tiff_tiled_separate_uint32(double ***graylevels, TIFF *tiffp, const uint32 tw, const uint32 th)
{
  unsigned int row,col,rr,cc;
  uint32 *buf = (uint32*)_TIFFmalloc(TIFFTileSize(tiffp));
  if (buf) {
    for (row = 0; row<ny; row+=th)
      for (col = 0; col<nx; col+=tw) {
	if (TIFFReadTile(tiffp,buf,col,row,0,0)<0) {_TIFFfree(buf); RAISE_ERROR("Invalid tile found into the TIFF image\n"); }
	uint32 *ptr = buf;
	for (rr = row; rr<MIN(row+th,ny); ++rr)
	  for (cc = col; cc<MIN(col+tw,nx); ++cc)
	    (**graylevels)[rr*nx+cc] = (double)(*(ptr++));
      }
    _TIFFfree(buf);
  }
  return EXIT_SUCCESS;
}

static int _load_tiff_tiled_separate_int32(double ***graylevels, TIFF *tiffp, const uint32 tw, const uint32 th)
{
  unsigned int row,col,rr,cc;
  int32 *buf = (int32*)_TIFFmalloc(TIFFTileSize(tiffp));
  if (buf) {
    for (row = 0; row<ny; row+=th)
      for (col = 0; col<nx; col+=tw) {
	if (TIFFReadTile(tiffp,buf,col,row,0,0)<0) {_TIFFfree(buf); RAISE_ERROR("Invalid tile found into the TIFF image\n"); }
	int32 *ptr = buf;
	for (rr = row; rr<MIN(row+th,ny); ++rr)
	  for (cc = col; cc<MIN(col+tw,nx); ++cc)
	    (**graylevels)[rr*nx+cc] = (double)(*(ptr++));
      }
    _TIFFfree(buf);
  }
  return EXIT_SUCCESS;
}

static int _load_tiff_tiled_separate_float(double ***graylevels, TIFF *tiffp, const uint32 tw, const uint32 th)
{
  unsigned int row,col,rr,cc;
  float *buf = (float*)_TIFFmalloc(TIFFTileSize(tiffp));
  if (buf) {
    for (row = 0; row<ny; row+=th)
      for (col = 0; col<nx; col+=tw) {
	if (TIFFReadTile(tiffp,buf,col,row,0,0)<0) {_TIFFfree(buf); RAISE_ERROR("Invalid tile found into the TIFF image\n"); }
	float *ptr = buf;
	for (rr = row; rr<MIN(row+th,ny); ++rr)
	  for (cc = col; cc<MIN(col+tw,nx); ++cc)
	    (**graylevels)[rr*nx+cc] = (double)(*(ptr++));
      }
    _TIFFfree(buf);
  }
  return EXIT_SUCCESS;
}


/***************************************************************************************
 * Striped Separate Planar Config = TIFF loaders for:
 *
 * 	- striped (=non-tiled) images, having separate planar config.
 * 	- supported types: uint8, int8, uint16, int16, uint32, int32 and float.
 * 	- multi-channel images supported
 *
 ***************************************************************************************/

static int _load_tiff_striped_separate_uint8(double ***graylevels, TIFF *tiffp)
{
  unsigned int rr,cc;
  uint8 *buf = (uint8*)_TIFFmalloc(TIFFStripSize(tiffp));
  if (buf) {
    uint32 row, rowsperstrip = (uint32)-1;
    TIFFGetField(tiffp,TIFFTAG_ROWSPERSTRIP,&rowsperstrip);
    for (row = 0; row<ny; row+= rowsperstrip) {
      uint32 nrow = (row+rowsperstrip>ny?ny-row:rowsperstrip);
      tstrip_t strip = TIFFComputeStrip(tiffp, row, 0);
      if ((TIFFReadEncodedStrip(tiffp,strip,buf,-1))<0) { _TIFFfree(buf); RAISE_ERROR("Invalid strip found into the TIFF image\n"); }
      uint8 *ptr = buf;
      for (rr = 0;rr<nrow; ++rr)
	for (cc = 0; cc<nx; ++cc)
	  (**graylevels)[(row+rr)*nx+cc] = (double)(*(ptr++));
    }
    _TIFFfree(buf);
  }
  return EXIT_SUCCESS;
}

static int _load_tiff_striped_separate_int8(double ***graylevels, TIFF *tiffp)
{
  unsigned int rr,cc;
  int8 *buf = (int8*)_TIFFmalloc(TIFFStripSize(tiffp));
  if (buf) {
    uint32 row, rowsperstrip = (uint32)-1;
    TIFFGetField(tiffp,TIFFTAG_ROWSPERSTRIP,&rowsperstrip);
    for (row = 0; row<ny; row+= rowsperstrip) {
      uint32 nrow = (row+rowsperstrip>ny?ny-row:rowsperstrip);
      tstrip_t strip = TIFFComputeStrip(tiffp, row, 0);
      if ((TIFFReadEncodedStrip(tiffp,strip,buf,-1))<0) { _TIFFfree(buf); RAISE_ERROR("Invalid strip found into the TIFF image\n"); }
      int8 *ptr = buf;
      for (rr = 0;rr<nrow; ++rr)
	for (cc = 0; cc<nx; ++cc)
	  (**graylevels)[(row+rr)*nx+cc] = (double)(*(ptr++));
    }
    _TIFFfree(buf);
  }
  return EXIT_SUCCESS;
}

static int _load_tiff_striped_separate_uint16(double ***graylevels, TIFF *tiffp)
{
  unsigned int rr,cc;
  uint16 *buf = (uint16*)_TIFFmalloc(TIFFStripSize(tiffp));
  if (buf) {
    uint32 row, rowsperstrip = (uint32)-1;
    TIFFGetField(tiffp,TIFFTAG_ROWSPERSTRIP,&rowsperstrip);
    for (row = 0; row<ny; row+= rowsperstrip) {
      uint32 nrow = (row+rowsperstrip>ny?ny-row:rowsperstrip);
      tstrip_t strip = TIFFComputeStrip(tiffp, row, 0);
      if ((TIFFReadEncodedStrip(tiffp,strip,buf,-1))<0) { _TIFFfree(buf); RAISE_ERROR("Invalid strip found into the TIFF image\n"); }
      uint16 *ptr = buf;
      for (rr = 0;rr<nrow; ++rr)
	for (cc = 0; cc<nx; ++cc)
	  (**graylevels)[(row+rr)*nx+cc] = (double)(*(ptr++));
    }
    _TIFFfree(buf);
  }
  return EXIT_SUCCESS;
}

static int _load_tiff_striped_separate_int16(double ***graylevels, TIFF *tiffp)
{
  unsigned int rr,cc;
  int16 *buf = (int16*)_TIFFmalloc(TIFFStripSize(tiffp));
  if (buf) {
    uint32 row, rowsperstrip = (uint32)-1;
    TIFFGetField(tiffp,TIFFTAG_ROWSPERSTRIP,&rowsperstrip);
    for (row = 0; row<ny; row+= rowsperstrip) {
      uint32 nrow = (row+rowsperstrip>ny?ny-row:rowsperstrip);
      tstrip_t strip = TIFFComputeStrip(tiffp, row, 0);
      if ((TIFFReadEncodedStrip(tiffp,strip,buf,-1))<0) { _TIFFfree(buf); RAISE_ERROR("Invalid strip found into the TIFF image\n"); }
      int16 *ptr = buf;
      for (rr = 0;rr<nrow; ++rr)
	for (cc = 0; cc<nx; ++cc)
	  (**graylevels)[(row+rr)*nx+cc] = (double)(*(ptr++));
    }
    _TIFFfree(buf);
  }
  return EXIT_SUCCESS;
}

static int _load_tiff_striped_separate_uint32(double ***graylevels, TIFF *tiffp)
{
  unsigned int rr,cc;
  uint32 *buf = (uint32*)_TIFFmalloc(TIFFStripSize(tiffp));
  if (buf) {
    uint32 row, rowsperstrip = (uint32)-1;
    TIFFGetField(tiffp,TIFFTAG_ROWSPERSTRIP,&rowsperstrip);
    for (row = 0; row<ny; row+= rowsperstrip) {
      uint32 nrow = (row+rowsperstrip>ny?ny-row:rowsperstrip);
      tstrip_t strip = TIFFComputeStrip(tiffp, row, 0);
      if ((TIFFReadEncodedStrip(tiffp,strip,buf,-1))<0) { _TIFFfree(buf); RAISE_ERROR("Invalid strip found into the TIFF image\n"); }
      uint32 *ptr = buf;
      for (rr = 0;rr<nrow; ++rr)
	for (cc = 0; cc<nx; ++cc)
	  (**graylevels)[(row+rr)*nx+cc] = (double)(*(ptr++));
    }
    _TIFFfree(buf);
  }
  return EXIT_SUCCESS;
}

static int _load_tiff_striped_separate_int32(double ***graylevels, TIFF *tiffp)
{
  unsigned int rr,cc;
  int32 *buf = (int32*)_TIFFmalloc(TIFFStripSize(tiffp));
  if (buf) {
    uint32 row, rowsperstrip = (uint32)-1;
    TIFFGetField(tiffp,TIFFTAG_ROWSPERSTRIP,&rowsperstrip);
    for (row = 0; row<ny; row+= rowsperstrip) {
      uint32 nrow = (row+rowsperstrip>ny?ny-row:rowsperstrip);
      tstrip_t strip = TIFFComputeStrip(tiffp, row, 0);
      if ((TIFFReadEncodedStrip(tiffp,strip,buf,-1))<0) { _TIFFfree(buf); RAISE_ERROR("Invalid strip found into the TIFF image\n"); }
      int32 *ptr = buf;
      for (rr = 0;rr<nrow; ++rr)
	for (cc = 0; cc<nx; ++cc)
	  (**graylevels)[(row+rr)*nx+cc] = (double)(*(ptr++));
    }
    _TIFFfree(buf);
  }
  return EXIT_SUCCESS;
}

static int _load_tiff_striped_separate_float(double ***graylevels, TIFF *tiffp)
{
  unsigned int rr,cc;
  float *buf = (float*)_TIFFmalloc(TIFFStripSize(tiffp));
  if (buf) {
    uint32 row, rowsperstrip = (uint32)-1;
    TIFFGetField(tiffp,TIFFTAG_ROWSPERSTRIP,&rowsperstrip);
    for (row = 0; row<ny; row+= rowsperstrip) {
      uint32 nrow = (row+rowsperstrip>ny?ny-row:rowsperstrip);
      tstrip_t strip = TIFFComputeStrip(tiffp, row, 0);
      if ((TIFFReadEncodedStrip(tiffp,strip,buf,-1))<0) { _TIFFfree(buf); RAISE_ERROR("Invalid strip found into the TIFF image\n"); }
      float *ptr = buf;
      for (rr = 0;rr<nrow; ++rr)
	for (cc = 0; cc<nx; ++cc)
	  (**graylevels)[(row+rr)*nx+cc] = (double)(*(ptr++));
    }
    _TIFFfree(buf);
  }
  return EXIT_SUCCESS;
}



/***************************************************************************************
 * Striped Contiguous Planar Config = TIFF loaders for:
 *
 * 	- striped (=non-tiled) images, having contiguous planar config.
 * 	- supported types: uint8, int8, uint16, int16, uint32, int32 and float.
 * 	- multi-channel images supported
 *
 ***************************************************************************************/

static int _load_tiff_striped_contig_uint8(double ***graylevels, TIFF *tiffp)
{
  unsigned int rr,cc;
  uint8 *buf = (uint8*)_TIFFmalloc(TIFFStripSize(tiffp));
  if (buf) {
    uint32 row, rowsperstrip = (uint32)-1;
    TIFFGetField(tiffp,TIFFTAG_ROWSPERSTRIP,&rowsperstrip);
    for (row = 0; row<ny; row+= rowsperstrip) {
      uint32 nrow = (row+rowsperstrip>ny?ny-row:rowsperstrip);
      tstrip_t strip = TIFFComputeStrip(tiffp, row, 0);
      if ((TIFFReadEncodedStrip(tiffp,strip,buf,-1))<0) { _TIFFfree(buf); RAISE_ERROR("Invalid strip found into the TIFF image\n"); }
      uint8 *ptr = buf;
      for (rr = 0; rr<nrow; ++rr)
	for (cc = 0; cc<nx; ++cc)
	  (**graylevels)[(row+rr)*nx+cc] = (double)(*(ptr++));
    }
    _TIFFfree(buf);
  }

  return EXIT_SUCCESS;
}

static int _load_tiff_striped_contig_int8(double ***graylevels, TIFF *tiffp)
{
  unsigned int rr,cc;
  int8 *buf = (int8*)_TIFFmalloc(TIFFStripSize(tiffp));
  if (buf) {
    uint32 row, rowsperstrip = (uint32)-1;
    TIFFGetField(tiffp,TIFFTAG_ROWSPERSTRIP,&rowsperstrip);
    for (row = 0; row<ny; row+= rowsperstrip) {
      uint32 nrow = (row+rowsperstrip>ny?ny-row:rowsperstrip);
      tstrip_t strip = TIFFComputeStrip(tiffp, row, 0);
      if ((TIFFReadEncodedStrip(tiffp,strip,buf,-1))<0) { _TIFFfree(buf); RAISE_ERROR("Invalid strip found into the TIFF image\n"); }
      int8 *ptr = buf;
      for (rr = 0; rr<nrow; ++rr)
	for (cc = 0; cc<nx; ++cc)
	  (**graylevels)[(row+rr)*nx+cc] = (double)(*(ptr++));
    }
    _TIFFfree(buf);
  }

  return EXIT_SUCCESS;
}

static int _load_tiff_striped_contig_uint16(double ***graylevels, TIFF *tiffp)
{
  unsigned int rr,cc;
  uint16 *buf = (uint16*)_TIFFmalloc(TIFFStripSize(tiffp));
  if (buf) {
    uint32 row, rowsperstrip = (uint32)-1;
    TIFFGetField(tiffp,TIFFTAG_ROWSPERSTRIP,&rowsperstrip);
    for (row = 0; row<ny; row+= rowsperstrip) {
      uint32 nrow = (row+rowsperstrip>ny?ny-row:rowsperstrip);
      tstrip_t strip = TIFFComputeStrip(tiffp, row, 0);
      if ((TIFFReadEncodedStrip(tiffp,strip,buf,-1))<0) { _TIFFfree(buf); RAISE_ERROR("Invalid strip found into the TIFF image\n"); }
      uint16 *ptr = buf;
      for (rr = 0; rr<nrow; ++rr)
	for (cc = 0; cc<nx; ++cc)
	  (**graylevels)[(row+rr)*nx+cc] = (double)(*(ptr++));
    }
    _TIFFfree(buf);
  }

  return EXIT_SUCCESS;
}


static int _load_tiff_striped_contig_int16(double ***graylevels, TIFF *tiffp)
{
  unsigned int rr,cc;
  int16 *buf = (int16*)_TIFFmalloc(TIFFStripSize(tiffp));
  if (buf) {
    uint32 row, rowsperstrip = (uint32)-1;
    TIFFGetField(tiffp,TIFFTAG_ROWSPERSTRIP,&rowsperstrip);
    for (row = 0; row<ny; row+= rowsperstrip) {
      uint32 nrow = (row+rowsperstrip>ny?ny-row:rowsperstrip);
      tstrip_t strip = TIFFComputeStrip(tiffp, row, 0);
      if ((TIFFReadEncodedStrip(tiffp,strip,buf,-1))<0) { _TIFFfree(buf); RAISE_ERROR("Invalid strip found into the TIFF image\n"); }
      int16 *ptr = buf;
      for (rr = 0; rr<nrow; ++rr)
	for (cc = 0; cc<nx; ++cc)
	  (**graylevels)[(row+rr)*nx+cc] = (double)(*(ptr++));
    }
    _TIFFfree(buf);
  }

  return EXIT_SUCCESS;
}

static int _load_tiff_striped_contig_uint32(double ***graylevels, TIFF *tiffp)
{
  unsigned int rr,cc;
  uint32 *buf = (uint32*)_TIFFmalloc(TIFFStripSize(tiffp));
  if (buf) {
    uint32 row, rowsperstrip = (uint32)-1;
    TIFFGetField(tiffp,TIFFTAG_ROWSPERSTRIP,&rowsperstrip);
    for (row = 0; row<ny; row+= rowsperstrip) {
      uint32 nrow = (row+rowsperstrip>ny?ny-row:rowsperstrip);
      tstrip_t strip = TIFFComputeStrip(tiffp, row, 0);
      if ((TIFFReadEncodedStrip(tiffp,strip,buf,-1))<0) { _TIFFfree(buf); RAISE_ERROR("Invalid strip found into the TIFF image\n"); }
      uint32 *ptr = buf;
      for (rr = 0; rr<nrow; ++rr)
	for (cc = 0; cc<nx; ++cc)
	  (**graylevels)[(row+rr)*nx+cc] = (double)(*(ptr++));
    }
    _TIFFfree(buf);
  }

  return EXIT_SUCCESS;
}


static int _load_tiff_striped_contig_int32(double ***graylevels, TIFF *tiffp)
{
  unsigned int rr,cc;
  int32 *buf = (int32*)_TIFFmalloc(TIFFStripSize(tiffp));
  if (buf) {
    uint32 row, rowsperstrip = (uint32)-1;
    TIFFGetField(tiffp,TIFFTAG_ROWSPERSTRIP,&rowsperstrip);
    for (row = 0; row<ny; row+= rowsperstrip) {
      uint32 nrow = (row+rowsperstrip>ny?ny-row:rowsperstrip);
      tstrip_t strip = TIFFComputeStrip(tiffp, row, 0);
      if ((TIFFReadEncodedStrip(tiffp,strip,buf,-1))<0) { _TIFFfree(buf); RAISE_ERROR("Invalid strip found into the TIFF image\n"); }
      int32 *ptr = buf;
      for (rr = 0; rr<nrow; ++rr)
	for (cc = 0; cc<nx; ++cc)
	  (**graylevels)[(row+rr)*nx+cc] = (double)(*(ptr++));
    }
    _TIFFfree(buf);
  }

  return EXIT_SUCCESS;
}

static int _load_tiff_striped_contig_float(double ***graylevels, TIFF *tiffp)
{
  unsigned int rr,cc;
  float *buf = (float*)_TIFFmalloc(TIFFStripSize(tiffp));
  if (buf) {
    uint32 row, rowsperstrip = (uint32)-1;
    TIFFGetField(tiffp,TIFFTAG_ROWSPERSTRIP,&rowsperstrip);
    for (row = 0; row<ny; row+= rowsperstrip) {
      uint32 nrow = (row+rowsperstrip>ny?ny-row:rowsperstrip);
      tstrip_t strip = TIFFComputeStrip(tiffp, row, 0);
      if ((TIFFReadEncodedStrip(tiffp,strip,buf,-1))<0) { _TIFFfree(buf); RAISE_ERROR("Invalid strip found into the TIFF image\n"); }
      float *ptr = buf;
      for (rr = 0; rr<nrow; ++rr)
	for (cc = 0; cc<nx; ++cc)
	  (**graylevels)[(row+rr)*nx+cc] = (double)(*(ptr++));
    }
    _TIFFfree(buf);
  }

  return EXIT_SUCCESS;
}
